package com.sun.messaging.jmq.jmsserver.persist.file;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.sun.messaging.jmq.io.Packet;
import com.sun.messaging.jmq.io.SysMessageID;
import com.sun.messaging.jmq.jmsserver.Globals;
import com.sun.messaging.jmq.jmsserver.config.BrokerConfig;
import com.sun.messaging.jmq.jmsserver.core.ConsumerUID;
import com.sun.messaging.jmq.jmsserver.core.Destination;
import com.sun.messaging.jmq.jmsserver.core.DestinationUID;
import com.sun.messaging.jmq.jmsserver.core.PacketReference;
import com.sun.messaging.jmq.jmsserver.core.Subscription;
import com.sun.messaging.jmq.jmsserver.data.BaseTransaction;
import com.sun.messaging.jmq.jmsserver.data.TransactionAcknowledgement;
import com.sun.messaging.jmq.jmsserver.data.TransactionState;
import com.sun.messaging.jmq.jmsserver.data.TransactionUID;
import com.sun.messaging.jmq.jmsserver.data.TransactionWork;
import com.sun.messaging.jmq.jmsserver.data.TransactionWorkMessage;
import com.sun.messaging.jmq.jmsserver.data.TransactionWorkMessageAck;
import com.sun.messaging.jmq.jmsserver.persist.Store;
import com.sun.messaging.jmq.jmsserver.resources.BrokerResources;
import com.sun.messaging.jmq.jmsserver.util.BrokerException;
import com.sun.messaging.jmq.jmsserver.util.lists.RemoveReason;
import com.sun.messaging.jmq.util.DestType;
import com.sun.messaging.jmq.util.SizeString;
import com.sun.messaging.jmq.util.lists.Reason;
import com.sun.messaging.jmq.util.log.Logger;
import com.sun.messaging.jmq.util.selector.SelectorFormatException;
import com.sun.messaging.jmq.util.txnlog.CheckPointListener;
import com.sun.messaging.jmq.util.txnlog.TransactionLogRecord;
import com.sun.messaging.jmq.util.txnlog.TransactionLogWriter;
import com.sun.messaging.jmq.util.txnlog.file.FileTransactionLogWriter;

/**
 * @author gsivewright
 * 
 * This class manages the transfer of data between the message store and the
 * transaction log.
 * 
 */

/*
 * Order of handling prepared txns
 * 
 * The following scenario describes a sequence of updates to the transaction log
 * and message store. This sequence is used to illustrate how consistent state
 * can be recreated on restart after a failure of the broker at various specific
 * processing points.
 * 
 * 1.write prepared txn to txnLog 2.checkpoint1 (see checkpoint sequence)
 * 3.write commit record to txnLog 4.write txn to store 5.write commit record to
 * preparedTxnStore 6.checkpoint2 (see checkpoint sequence)
 * 
 * checkpoint sequence:
 * 
 * 1.sync message store 2.write prepared txns (since last checkpoint) to
 * prepared store 3.sync prepareTxnStore 4.reset txn log 5.remove committed txns
 * (since last checkpoint) from prepared store
 * 
 * 
 * replay after failure at 1 > 2.2
 * 
 * 1.read preparedTxnStore (empty) 2.read prepared txn from txnLog 3.check if
 * txn is in preparedTxnStore. (add). 4.checkpoint
 * 
 * replay after failure at 2.3 -> 2.4
 * 
 * 1.read txns from preparedTxnStore 2.read prepared txn from txnLog 3.check if
 * txn is in preparedTxnStore. (already there) 4.checkpoint
 * 
 * replay after failure at 2.5
 * 
 * 1.read preparedTxn from preparedTxnStore 2.checkpoint
 * 
 * replay after failure at 3 -> 5
 * 
 * 1.read txns from preparedTxnStore 2.read commit record from txnLog 3.write
 * txn to message store (replace if there) 4.write commit record to
 * preparedTxnStore (if not committed) 5.checkpoint
 * 
 * replay after failure at 6.4 1)read txns from preparedTxnStore (find committed
 * txn) 2)do not replay (not in reset txnLog) 3)checkpoint
 * 
 */

public class TransactionLogManager implements CheckPointListener {

	static final String TXNLOG_PROP_PREFIX = Globals.IMQ
			+ ".persist.file.txnLog";
	static final String TXNLOG_FILE_SIZE_PROP = TXNLOG_PROP_PREFIX
			+ ".file.size";
	static final String SYNC_TXNLOG_WRITES_PROP = TXNLOG_PROP_PREFIX
			+ ".syncWrites";

	static final long DEFAULT_TXNLOG_FILE_SIZE = 1024 * 10; // 10M

	static final String MSG_LOG_FILENAME = "txnlog";

	public static final BrokerResources br = Globals.getBrokerResources();
	Store store;
	MsgStore msgStore;
	File rootDir;

	private TransactionLogWriter msgLogWriter = null;

	LocalTransactionManager localTransactionManager;
	ClusterTransactionManager clusterTransactionManager;
	RemoteTransactionManager remoteTransactionManager;

	CheckpointManager checkpointManager;
	PreparedTxnStore preparedTxnStore;

	public static final Logger logger = Globals.getLogger();

	public TransactionLogManager(Store store,
			PreparedTxnStore preparedTxnStore, MsgStore msgStore, File rootDir) {

		this.store = store;
		this.preparedTxnStore = preparedTxnStore;
		this.msgStore = msgStore;
		this.rootDir = rootDir;
		this.localTransactionManager = new LocalTransactionManager(this);
		this.clusterTransactionManager = new ClusterTransactionManager(this);
		this.remoteTransactionManager = new RemoteTransactionManager(this);
		this.checkpointManager = new CheckpointManager(this);
	}

	public void startup() throws BrokerException {
		// These are the stages on broker startup
		// 1) read in any prepared transactions held in preparedTxnStore.
		// 2) check if the transaction log needs replaying.
		// 3) replay committed transaction to message store
		// 4) after replay is complete, add any remaining transactions to
		// preparedTxnStore.
		// 5) sync message store.
		// 6) reset transaction log.
		// 7) remove any committed transactions from preparedTxnStore

		if (Store.getDEBUG()) {
			logger.log(Logger.DEBUG, getPrefix() + " startup");
		}
		// 1 read in any prepared transactions held in preparedTxnStore.
		processStoredTxnsOnStartup();

		// 2) check if the transaction log needs replaying.
		initTransactionLogOnStartUp();

		// 3) replay committed transaction to message store
		replayTransactionLogOnStartup();

		// 4) after replay is complete, add any remaining transactions to
		// preparedTxnStore.

		// 5) checkpoint message store.
		store.syncDestination(null);

		try {
			// 6) reset transaction log.
			msgLogWriter.reset();
		} catch (IOException e) {
			throw new BrokerException("failed to reset transaction log", e);
		}

		// 7) remove any committed transactions from preparedTxnStore
		removeCommittedTransactionsOnStartup();

	}

	void processStoredTxnsOnStartup() throws BrokerException {
		if (Store.getDEBUG()) {
			String msg = getPrefix() + " processStoredTxnsOnStartup";
			logger.log(Logger.DEBUG, msg);
		}

		preparedTxnStore.loadTransactions();

		// for now just read and sort into different types
		Enumeration<BaseTransaction> transactions = preparedTxnStore
				.txnEnumeration();
		while (transactions.hasMoreElements()) {
			BaseTransaction baseTxn = transactions.nextElement();
			// need to switch on all types
			int type = baseTxn.getType();
			BaseTransactionManager baseTxnMan = this
					.getTransactionManager(type);
			baseTxnMan.processStoredTxnOnStartup(baseTxn);
		}

	}

	void removeCommittedTransactionsOnStartup() throws BrokerException {

		if (Store.getDEBUG()) {
			String msg = getPrefix()
					+ " removeCommitedTransactionsInPreparedTxnStore";
			logger.log(Logger.DEBUG, msg);
		}
		// for now just report on any transactions
		Enumeration<BaseTransaction> transactions = preparedTxnStore
				.txnEnumeration();
		List<BaseTransaction> committed = new ArrayList<BaseTransaction>();
		while (transactions.hasMoreElements()) {
			BaseTransaction baseTxn = transactions.nextElement();
			if (baseTxn.getState() == TransactionState.COMMITTED) {
				committed.add(baseTxn);
			}

		}
		Iterator<BaseTransaction> iter = committed.iterator();
		while (iter.hasNext()) {
			BaseTransaction baseTxn = iter.next();
			TransactionUID tid = baseTxn.getTid();
			preparedTxnStore.removeTransaction(tid, true);
			if (Store.getDEBUG()) {
				String msg = getPrefix()
						+ " removed committed transaction from preparedTxnStore. Tid="
						+ tid;
				logger.log(Logger.DEBUG, msg);
			}
		}

	}
	public void resetTransactionLogOnStartUp() throws BrokerException {
		logger.log(logger.INFO, "resetting txn Log");
		String filename = MSG_LOG_FILENAME;
		File file = new File(rootDir,filename);
		file.delete();
	}

	public void initTransactionLogOnStartUp() throws BrokerException {

		if (Store.getDEBUG()) {
			String msg = getPrefix() + " initTransactionLogOnStartUp";
			logger.log(Logger.DEBUG, msg);
		}
		logger.log(logger.INFO, BrokerResources.I_TXNLOG_ENABLED);

		// create txn log writers
		String filename = null;
		try {
			BrokerConfig config = Globals.getConfig();
			SizeString filesize = config.getSizeProperty(TXNLOG_FILE_SIZE_PROP,
					DEFAULT_TXNLOG_FILE_SIZE);

			boolean syncAllTxnWrites = config.getBooleanProperty(
					SYNC_TXNLOG_WRITES_PROP, true);
			if (Store.getDEBUG()) {
				logger.log(Logger.DEBUG, "syncAllTxnWrites = "
						+ syncAllTxnWrites);
			}
			filename = MSG_LOG_FILENAME;

			String mode = "rwd";
			boolean synch = true;
			if (!syncAllTxnWrites) {
				mode = "rw";
				synch = false;
			}
			logger.log(logger.INFO, "mode=" + mode);

			msgLogWriter = new FileTransactionLogWriter(rootDir, filename,
					filesize.getBytes(), mode, synch);
			msgLogWriter.setCheckPointListener(this);
			if (Store.getDEBUG()) {
				logger.log(Logger.DEBUG, "created txn log");
			}
		} catch (IOException ex) {
			logger.logStack(Logger.ERROR,
					BrokerResources.E_CREATE_TXNLOG_FILE_FAILED, filename, ex);
			throw new BrokerException(br.getString(
					BrokerResources.E_CREATE_TXNLOG_FILE_FAILED, filename), ex);
		}

	}

	public void replayTransactionLogOnStartup() throws BrokerException {

		if (Store.getDEBUG()) {
			logger.log(Logger.DEBUG, getPrefix()
					+ " replayTransactionLogOnStartup");
		}

		try {
			if (msgLogWriter.playBackRequired()) {
				if (Store.getDEBUG()) {
					String msg = getPrefix()
							+ " replayTransactionLogOnStartup: playBackRequired";
					logger.log(Logger.DEBUG, msg);
				}

				logger.log(Logger.FORCE, BrokerResources.I_PROCESS_MSG_TXNLOG);

				// All destinations need to be loaded
				Destination.init();
				Subscription.initSubscriptions();
				HashSet dstLoadedSet = new HashSet(); // Keep track of loaded
				// dst

				int count = 0;

				Map<TransactionUID, BaseTransaction> inProgressTransactions = new HashMap<TransactionUID, BaseTransaction>();

				Iterator<TransactionLogRecord> itr = msgLogWriter.iterator();
				while (itr.hasNext()) {
					count++; // Keep track the number of records processed

					// Read in the acks or msgs & acks
					TransactionLogRecord rec = (TransactionLogRecord) itr
							.next();

					byte[] data = rec.getBody();
					TransactionEvent txnEvent = readTransactionEvent(data);
					int type = txnEvent.getType();
					if (Store.getDEBUG()) {
						String msg = getPrefix()
								+ " replayTransactionLogOnStartup() read "
								+ txnEvent;
						logger.log(Logger.DEBUG, msg);
					}

					BaseTransactionManager tm = getTransactionManager(type);
					tm.replayTransactionEvent(txnEvent, dstLoadedSet);

				}

			} else {
				if (Store.getDEBUG()) {
					logger.log(Logger.DEBUG, "no playBackRequired");
				}
			}
		}

		catch (IOException e) {
			e.printStackTrace();

		}
	}

	void replayTransactionWork(TransactionWork txnWork, TransactionUID tid,
			Set dstLoadedSet) throws IOException, BrokerException {

		if (Store.getDEBUG()) {
			String msg = getPrefix() + " replayTransactionWork for tid " + tid
					+ " txnWork=" + txnWork;
			logger.log(Logger.DEBUG, msg);
		}
		replaySentMessages(txnWork.getSentMessages(), dstLoadedSet);
		replayAcknowledgedMessages(txnWork.getMessageAcknowledgments(),
				dstLoadedSet);

	}

	TransactionEvent readTransactionEvent(byte[] data) throws IOException,
			BrokerException {
		TransactionEvent event = TransactionEvent.createFromBytes(data);
		return event;
	}

	private void replaySentMessages(List<TransactionWorkMessage> sentMessages,
			Set dstLoadedSet) throws IOException, BrokerException {

		if (Store.getDEBUG()) {
			logger.log(Logger.DEBUG, getPrefix() + " replaySentMessages");
		}

		for (int i = 0; i < sentMessages.size(); i++) {
			TransactionWorkMessage workMessage = sentMessages.get(i);

			// Reconstruct the message
			Packet pkt = workMessage.getMessage();

			SysMessageID mid = pkt.getSysMessageID();
			if (Store.getDEBUG()) {
				String msg = getPrefix() + " replaying sent message: "
						+ workMessage;
				logger.log(Logger.DEBUG, msg);
			}
			// Make sure dst exists; autocreate if possible
			Destination dst = Destination.getDestination(pkt.getDestination(),
					pkt.getIsQueue() ? DestType.DEST_TYPE_QUEUE
							: DestType.DEST_TYPE_TOPIC, true, true);
			DestinationUID did = dst.getDestinationUID();

			// Load all msgs inorder to verify if any msgs are missing
			if (!dstLoadedSet.contains(dst)) {
				dst.load();
				dstLoadedSet.add(dst); // Keep track of what has been loaded
			}

			// Check to see if the msg is in the store

			boolean exists = msgStore.containsMessage(did, mid);
			if (exists) {
				logger.log(Logger.FORCE, BrokerResources.I_REPLACE_MSG_TXNLOG,
						mid);
				msgStore.removeMessage(did, mid, false);
			} else {
				logger.log(Logger.FORCE,
						BrokerResources.I_RECONSTRUCT_MSG_TXNLOG, mid);
			}

			PacketReference pr = PacketReference
					.createReference(pkt, did, null);

			try {
				dst.routeNewMessage(pr);
			} catch (SelectorFormatException e) {
				// shouldn't happens
				throw new BrokerException(br.getString(
						BrokerResources.E_ROUTE_RECONSTRUCTED_MSG_FAILED, mid),
						e);
			}
		}
	}

	private void replayAcknowledgedMessages(
			List<TransactionWorkMessageAck> acknowledgedMessages,
			Set dstLoadedSet) throws IOException, BrokerException {

		for (int i = 0; i < acknowledgedMessages.size(); i++) {
			TransactionWorkMessageAck messageAck = acknowledgedMessages.get(i);

			String name = messageAck.getDest(); // Destination ID
			DestinationUID did = new DestinationUID(name);
			SysMessageID mid = messageAck.getSysMessageID();
			ConsumerUID iid = messageAck.getConsumerID();
			if (Store.getDEBUG()) {
				String msg = getPrefix() + " replaying acknowledged message "
						+ messageAck;
				logger.log(Logger.DEBUG, msg);
			}
			// Make sure dst exists; autocreate if possible
			Destination dst = Destination.getDestination(did.getName(), did
					.isQueue() ? DestType.DEST_TYPE_QUEUE
					: DestType.DEST_TYPE_TOPIC, true, true);

			// Load all msgs inorder to update consumer states
			if (!dstLoadedSet.contains(dst)) {
				dst.load();
				dstLoadedSet.add(dst); // Keep track of what has been loaded
			}

			if (msgStore.containsMessage(did, mid)) {
				logger.log(logger.FORCE,
						BrokerResources.I_UPDATE_INT_STATE_TXNLOG, iid, mid);
				// For Queue, ensure the stored ConsumerUID is 0 otherwise
				// use try using the correct value; see bug 6516160
				if (dst.isQueue() && iid.longValue() != 0) {
					msgStore.updateInterestState(did, mid, PacketReference
							.getQueueUID(), Store.INTEREST_STATE_ACKNOWLEDGED,
							false);
				} else {
					msgStore.updateInterestState(did, mid, iid,
							Store.INTEREST_STATE_ACKNOWLEDGED, false);
				}

				acknowledgeOnReplay(dst,mid,iid);
				
				

			} else {
				logger.log(logger.FORCE,
						BrokerResources.I_DISREGARD_INT_STATE_TXNLOG, iid, mid);
			}
		}
	}

	
	public void acknowledgeOnReplay(Destination dst, SysMessageID mid,
			ConsumerUID iid) throws IOException, BrokerException {
		PacketReference ref = dst.getMessage(mid);
		boolean allAcked = false;
		if (ref != null) {
			allAcked = ref.acknowledgedOnReplay(iid, iid);
			if (Store.getDEBUG()) {
				String msg = getPrefix() + " acknowledgedOnReplay  " + mid
						+ " allAcked =" + allAcked;
				logger.log(Logger.DEBUG, msg);
			}
		} else {
			if (Store.getDEBUG()) {
				String msg = getPrefix()
						+ " did not find packet in destination " + mid;
				logger.log(Logger.DEBUG, msg);
			}
		}
		if (allAcked) {
			dst.removeMessage(mid, RemoveReason.ACKNOWLEDGED);
			if (Store.getDEBUG()) {
				String msg = getPrefix() + " removed message from destination "
						+ mid;
				logger.log(Logger.DEBUG, msg);
			}
		}
	}
	
	public void replayRemoteAcks(TransactionAcknowledgement[] txnAcks,
			TransactionUID tid, HashSet dstLoadedSet) throws IOException,BrokerException {
		// TO DO
		// store destUID alongside acks so that we can replay more directly

		if (Store.getDEBUG()) {
			String msg = getPrefix() + " replayRemoteAcks ";
			logger.log(Logger.DEBUG, msg);
		}
		for (int i = 0; i < txnAcks.length; i++) {
			TransactionAcknowledgement txnAck = txnAcks[i];
			SysMessageID mid = txnAck.getSysMessageID();
			ConsumerUID iid = txnAck.getStoredConsumerUID();

			PacketReference pr = Destination.get(mid);
			if (pr == null) {
				// could have been acknowledged already?
				// maybe thedestination has not been loaded

				// TO DO check this further
				String msg = " could not find packet for replayed message ack "
						+ txnAck + " in transaction " + tid;
				logger.log(Logger.WARNING, msg);
			} else {
				Destination dst = pr.getDestination();
				DestinationUID did = dst.getDestinationUID();
				if (msgStore.containsMessage(did, mid)) {
					logger
							.log(logger.FORCE,
									BrokerResources.I_UPDATE_INT_STATE_TXNLOG,
									iid, mid);
					// For Queue, ensure the stored ConsumerUID is 0 otherwise
					// use try using the correct value; see bug 6516160
					if (dst.isQueue() && iid.longValue() != 0) {
						msgStore.updateInterestState(did, mid, PacketReference
								.getQueueUID(),
								Store.INTEREST_STATE_ACKNOWLEDGED, false);
					} else {
						msgStore.updateInterestState(did, mid, iid,
								Store.INTEREST_STATE_ACKNOWLEDGED, false);
					}
					
					acknowledgeOnReplay(dst,mid,iid);
				} else {
					logger.log(logger.FORCE,
							BrokerResources.I_DISREGARD_INT_STATE_TXNLOG, iid,
							mid);
				}
			}

		}

	}

	public void loggedCommitWrittenToMessageStore(TransactionUID tid, int type) {
		BaseTransactionManager txnMan = getTransactionManager(type);

		txnMan.playingToMessageStoreComplete(tid);

	}

	BaseTransactionManager getTransactionManager(int type) {
		BaseTransactionManager r = null;
		switch (type) {
		case BaseTransaction.LOCAL_TRANSACTION_TYPE:
			r = localTransactionManager;
			break;
		case BaseTransaction.CLUSTER_TRANSACTION_TYPE:
			r = clusterTransactionManager;
			break;
		case BaseTransaction.REMOTE_TRANSACTION_TYPE:
			r = remoteTransactionManager;
			break;
		case BaseTransaction.UNDEFINED_TRANSACTION_TYPE:
			throw new UnsupportedOperationException(
					"UNDEFINED_TRANSACTION_TYPE");
		default:
			throw new UnsupportedOperationException("unknown type:" + type);
		}
		return r;
	}

	public void logTxn(BaseTransaction baseTxn) throws BrokerException {
		if (Store.getDEBUG()) {
			logger.log(Logger.DEBUG, getPrefix() + " logTxn() " + baseTxn);
		}

		try {
			store.txnLogSharedLock.lock();
			int type = baseTxn.getType();
			BaseTransactionManager txnManager = getTransactionManager(type);
			txnManager.processTxn(baseTxn);
			TransactionEvent txnEvent = txnManager
					.generateEvent(baseTxn, false);

			byte[] data = txnEvent.writeToBytes();

			TransactionLogRecord record = msgLogWriter
					.newTransactionLogRecord();
			// record.setType(txnWork.getType());
			record.setBody(data);
			msgLogWriter.write(record);
			// baseTxn.setLogged(true);
		} catch (IOException ioe) {
			throw new BrokerException("error logging transaction", ioe);
		} finally {
			store.txnLogSharedLock.unlock();
		}
	}

	public void logTxnCompletion(TransactionUID tid, int completionState,
			int type) throws BrokerException {
		if (Store.getDEBUG()) {
			logger
					.log(Logger.DEBUG, getPrefix() + " logTxnCompletion() "
							+ tid);
		}
		try {
			store.txnLogSharedLock.lock();
			BaseTransactionManager txnMan = getTransactionManager(type);
			BaseTransaction existing = txnMan.processTxnCompletion(tid,
					completionState);
			TransactionEvent txnEvent = txnMan.generateEvent(existing, true);

			byte[] data = txnEvent.writeToBytes();

			TransactionLogRecord record = msgLogWriter
					.newTransactionLogRecord();
			// record.setType(txnWork.getType());
			record.setBody(data);
			msgLogWriter.write(record);

		} catch (IOException ioe) {
			throw new BrokerException("error in logTxnCompletion", ioe);
		} finally {
			store.txnLogSharedLock.unlock();
		}
	}

	public void doCheckpoint() {

		if (Store.getDEBUG()) {
			logger.log(Logger.DEBUG, getPrefix() + " doCheckpoint");
		}
		// 1. get exclusive lock on transaction log
		// 2. wait for all logged commits to be written to message store
		// 3. sync message store
		// 4. write prepared txns (since last checkpoint) to prepared store
		// 5. sync prepareTxnStore
		// 6. reset txn log
		// 7. remove committed txns (since last checkpoint) from prepared store

		Store store = null;
		try {
			store = Globals.getStore();
		} catch (Throwable e) {
			logger.logStack(Logger.ERROR, "failed to getStore", e);

			// fatal
		}

		try {

			// 1. get an exclusive lock on txnLog to prevent any other updates
			// occurring
			store.txnLogExclusiveLock.lock();

			// 2. wait for all logged commits to be written to message store
			localTransactionManager.waitForPlayingToMessageStoreCompletion();
			clusterTransactionManager.waitForPlayingToMessageStoreCompletion();
			remoteTransactionManager.waitForPlayingToMessageStoreCompletion();

			// 3. sync the message store
			store.syncStore();

			// 4. write prepared transactions to prepared transaction store
			localTransactionManager
					.writePreparedTransactionsToPreparedTxnStoreOnCheckpoint();
			clusterTransactionManager
					.writePreparedTransactionsToPreparedTxnStoreOnCheckpoint();
			remoteTransactionManager
					.writePreparedTransactionsToPreparedTxnStoreOnCheckpoint();

			// 5. sync prepareTxnStore
			preparedTxnStore.sync();

			// 6. reset txn log
			msgLogWriter.checkpoint();

			// 7. remove committed txns (since last checkpoint) from prepared
			// store
			localTransactionManager.removeCompleteTransactionsAfterCheckpoint();
			clusterTransactionManager
					.removeCompleteTransactionsAfterCheckpoint();
			remoteTransactionManager
					.removeCompleteTransactionsAfterCheckpoint();

		} catch (Throwable e) {
			String msg = getPrefix()
					+ "Failed to synchronize persistence store for transaction log checkpoint";
			logger.logStack(Logger.ERROR,
					BrokerResources.E_INTERNAL_BROKER_ERROR, msg, e);
		} finally {
			store.txnLogExclusiveLock.unlock();

			if (Store.getDEBUG()) {
				logger
						.log(Logger.DEBUG, getPrefix()
								+ " doCheckpoint complete");
			}
		}
	}

	// checkpoint handling

	public final void checkpoint() {
		if (Store.getDEBUG()) {
			logger.log(Logger.DEBUG, getPrefix() + " request a checkpoint");
		}

		checkpointManager.enqueueCheckpoint();
	}

	String getPrefix() {
		return "TransactionLogManager: " + Thread.currentThread().getName();
	}

	public LocalTransactionManager getLocalTransactionManager() {
		return localTransactionManager;
	}

	public ClusterTransactionManager getClusterTransactionManager() {
		return clusterTransactionManager;
	}

	public RemoteTransactionManager getRemoteTransactionManager() {
		return remoteTransactionManager;
	}
}
